/////////////////////////////////////////////////////////////////////
//
//            X   X           X
//           XXX XX         XX
//          XXXXXXXX      XXX
//         XX X XXXXXXXXXXX
//        XXXXX XXXXXXXXX
//       XXXXX XXXXXXXXXX
//            XXX XXX XXX
//           XXX XX   XX
//           X   X     X
//
//    Copyright (C) 2003-2022  Ron Jakl
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////


#include <QApplication>
#include <QGridLayout>
#include <QLabel>
#include <QLineEdit>
#include <QSpacerItem>
#include <QWidget>
#include <QTimer>
#include <QPixmap>
#include <QResizeEvent>
#include <QScreen>
#include <QSize>
#include <QRect>
#include <QFont>
#include <cmath>
#include <assert.h>

#include "accelerationcalculator.h"

TAccelerationCalculator::TAccelerationCalculator(
	const QWidget						*parent,
	const Qt::WindowFlags				flags)
:QMainWindow(const_cast<QWidget*>(parent),flags)
{
	QScreen								*screen;
	QLabel								*acceleration_label;
    QLabel								*acceleration_units_label;
	QLabel								*initial_velocity_label;
    QLabel								*initial_velicity_units_label;
	QLabel								*final_velocity_label;
	QLabel								*final_velocity_units_label;
    QLabel								*transition_time_label;
	QLabel								*transition_time_units_label;
	QLabel								*transition_distance_label;
	QLabel								*transition_distance_units_label;
    QSpacerItem							*control_upper_spacer;
    QSpacerItem							*control_lower_spacer;
	QFont								copyright_font;

	this->resize(901,401);
	
	copyright_font.setFamily(QStringLiteral("Verdana"));
	copyright_font.setPointSize(8);
	copyright_font.setItalic(true);

    d_central_widget = new QWidget(this);
    this->setCentralWidget(d_central_widget);

    d_central_widget_layout = new QGridLayout(d_central_widget);

	d_control_layout = new QGridLayout();

    control_upper_spacer = new QSpacerItem(0,0,QSizePolicy::Minimum,QSizePolicy::Expanding);
    d_control_layout->addItem(control_upper_spacer,0,0,1,3);

    initial_velocity_label = new QLabel(d_central_widget);
	initial_velocity_label->setAlignment(Qt::AlignRight|Qt::AlignTrailing|Qt::AlignVCenter);
    d_control_layout->addWidget(initial_velocity_label,1,0,1,1);

    d_initial_velocity_edit = new QLineEdit(d_central_widget);
	d_initial_velocity_edit->setAlignment(Qt::AlignCenter);
    d_control_layout->addWidget(d_initial_velocity_edit,1,1,1,1);

    initial_velicity_units_label = new QLabel(d_central_widget);
    d_control_layout->addWidget(initial_velicity_units_label,1,2,1,1);

    final_velocity_label = new QLabel(d_central_widget);
	final_velocity_label->setAlignment(Qt::AlignRight|Qt::AlignTrailing|Qt::AlignVCenter);
    d_control_layout->addWidget(final_velocity_label,2,0,1,1);

    d_final_velocity_edit = new QLineEdit(d_central_widget);
	d_final_velocity_edit->setAlignment(Qt::AlignCenter);
    d_control_layout->addWidget(d_final_velocity_edit,2,1,1,1);

    final_velocity_units_label = new QLabel(d_central_widget);
    d_control_layout->addWidget(final_velocity_units_label,2,2,1,1);

    acceleration_label = new QLabel(d_central_widget);
	acceleration_label->setAlignment(Qt::AlignRight|Qt::AlignTrailing|Qt::AlignVCenter);
    d_control_layout->addWidget(acceleration_label,3,0,1,1);

    d_acceleration_edit = new QLineEdit(d_central_widget);
	d_acceleration_edit->setAlignment(Qt::AlignCenter);
    d_control_layout->addWidget(d_acceleration_edit,3,1,1,1);

    acceleration_units_label = new QLabel(d_central_widget);
    d_control_layout->addWidget(acceleration_units_label,3,2,1,1);

    transition_time_label = new QLabel(d_central_widget);
	transition_time_label->setAlignment(Qt::AlignRight|Qt::AlignTrailing|Qt::AlignVCenter);
    d_control_layout->addWidget(transition_time_label,4,0,1,1);

    d_transition_time_edit = new QLineEdit(d_central_widget);
	d_transition_time_edit->setAlignment(Qt::AlignCenter);
	d_transition_time_edit->setReadOnly(true);
    d_control_layout->addWidget(d_transition_time_edit,4,1,1,1);

    transition_time_units_label = new QLabel(d_central_widget);
    d_control_layout->addWidget(transition_time_units_label,4,2,1,1);
	
    transition_distance_label = new QLabel(d_central_widget);
	transition_distance_label->setAlignment(Qt::AlignRight|Qt::AlignTrailing|Qt::AlignVCenter);
    d_control_layout->addWidget(transition_distance_label,5,0,1,1);
	
    d_transition_distance_edit = new QLineEdit(d_central_widget);
	d_transition_distance_edit->setAlignment(Qt::AlignCenter);
	d_transition_distance_edit->setReadOnly(true);
    d_control_layout->addWidget(d_transition_distance_edit,5,1,1,1);
	
    transition_distance_units_label = new QLabel(d_central_widget);
    d_control_layout->addWidget(transition_distance_units_label,5,2,1,1);

    control_lower_spacer = new QSpacerItem(0,0,QSizePolicy::Minimum,QSizePolicy::Expanding);
    d_control_layout->addItem(control_lower_spacer,6,0,1,3);
    d_central_widget_layout->addLayout(d_control_layout,0,0,1,1);

    d_graph_label = new QLabel(d_central_widget);
    d_graph_label->setFrameShape(QFrame::StyledPanel);
    d_graph_label->setAlignment(Qt::AlignCenter);
//    d_graph_label->setAlignment(Qt::AlignHCenter|Qt::AlignBottom);
//    d_graph_label->setScaledContents(true);
    d_central_widget_layout->addWidget(d_graph_label,0,1,1,1);
	
	d_copyright_label = new QLabel(this);
	d_copyright_label->setFont(copyright_font);
	d_central_widget_layout->addWidget(d_copyright_label,1,0,1,2);	// assuming wide startup
	
	d_update_timer = new QTimer(this);
	d_update_timer->setSingleShot(true);
	
	// defaults
	d_display_mode = TAccelerationCalculator::DISPLAY_WIDE;
    d_display_graph_name = QStringLiteral(":/graphs/graph_up.png");
	d_graph_label->setPixmap(QPixmap(d_display_graph_name));

	this->setWindowTitle(QStringLiteral("Acceleration Calculator"));
	
	initial_velocity_label->setText(QStringLiteral("Initial Velocity:"));
	initial_velicity_units_label->setText(QStringLiteral("mm/sec"));
	final_velocity_label->setText(QStringLiteral("Final Velocity:"));
	final_velocity_units_label->setText(QStringLiteral("mm/sec"));
	acceleration_label->setText(QStringLiteral("Acceleration:"));
	acceleration_units_label->setText(QStringLiteral("mm/sec^2"));
	transition_time_label->setText(QStringLiteral("Transition Time:"));
	transition_time_units_label->setText(QStringLiteral("sec"));
	transition_distance_label->setText(QStringLiteral("Distance:"));
	transition_distance_units_label->setText(QStringLiteral("mm"));

	d_initial_velocity_edit->setText(QStringLiteral("0.0"));
	d_final_velocity_edit->setText(QStringLiteral("6.0"));
	d_acceleration_edit->setText(QStringLiteral("30.0"));
	d_transition_time_edit->setText(QStringLiteral("0.200"));
	d_transition_distance_edit->setText(QStringLiteral("0.600"));
	d_copyright_label->setText(QStringLiteral("Copyright (C) 2022 Select Calibration Incorporated.  www.selectcalibration.ca"));

	screen = QGuiApplication::primaryScreen();
	screen->setOrientationUpdateMask(Qt::PortraitOrientation | Qt::LandscapeOrientation | Qt::InvertedPortraitOrientation | Qt::InvertedLandscapeOrientation);

	connect(screen,&QScreen::orientationChanged,this,&TAccelerationCalculator::Screen_Orientation_Changed);
	connect(d_initial_velocity_edit,&QLineEdit::textChanged,this,&TAccelerationCalculator::Input_Text_Changed);
	connect(d_final_velocity_edit,&QLineEdit::textChanged,this,&TAccelerationCalculator::Input_Text_Changed);
	connect(d_acceleration_edit,&QLineEdit::textChanged,this,&TAccelerationCalculator::Input_Text_Changed);
	connect(d_update_timer,&QTimer::timeout,this,&TAccelerationCalculator::Update_Fields);

    this->Screen_Orientation_Changed(screen->orientation());
}

TAccelerationCalculator::~TAccelerationCalculator(void)
{
}

void TAccelerationCalculator::Screen_Orientation_Changed(
	Qt::ScreenOrientation				orientation)
{
	if(orientation == Qt::LandscapeOrientation ||
	   orientation == Qt::InvertedLandscapeOrientation)
	{
		if(d_display_mode != TAccelerationCalculator::DISPLAY_WIDE)
		{
			d_display_mode = TAccelerationCalculator::DISPLAY_WIDE;
			this->Update_Layout();
		}
	}
	else if(orientation == Qt::PortraitOrientation ||
			orientation == Qt::InvertedPortraitOrientation)
	{
		if(d_display_mode != TAccelerationCalculator::DISPLAY_TALL)
		{
			d_display_mode = TAccelerationCalculator::DISPLAY_TALL;
			this->Update_Layout();
		}
	}
	
	this->Update_Graph(this->rect().size());
}

void TAccelerationCalculator::Input_Text_Changed(void)
{
	static const int					CALCULATION_DELAY_TIME(500);	// time before input fields updated
	
	d_update_timer->start(CALCULATION_DELAY_TIME);
}

void TAccelerationCalculator::Update_Fields(void)
{
	double								initial_velocity;
	double								final_velocity;
	double								velocity;
	double								acceleration;
	double								time;
	double								distance;
	bool								state;
	static const double					ZERO_EPSILON(0.001);

	
	initial_velocity = d_initial_velocity_edit->text().toDouble(&state);
	if(!state) return;
	
	final_velocity = d_final_velocity_edit->text().toDouble(&state);
	if(!state) return;
	
	acceleration = d_acceleration_edit->text().toDouble(&state);
	if(!state) return;
	
	velocity = fabs(final_velocity - initial_velocity);
	
	if(velocity < ZERO_EPSILON || fabs(acceleration) < ZERO_EPSILON)
	{
		d_transition_time_edit->setText(QString());
		d_transition_distance_edit->setText(QString());
	}
	else
	{
		
		// t = v/a
		time = velocity / acceleration;
		
		d_transition_time_edit->setText(QString("%1").arg(time,0,'f',3));
		
		
		// distance = 1/2 * at^2
		distance = acceleration * time * time;
		distance /= 2;
		
		d_transition_distance_edit->setText(QString("%1").arg(distance,0,'f',3));
	}

	this->Update_Graph(this->rect().size());
	
}

void TAccelerationCalculator::resizeEvent(
	QResizeEvent						*event)
{
	QSize								size(event->size());
	
	QMainWindow::resizeEvent(event);
	
	if(size.width() > 0 && size.height() > 0)
	{
		if(size.width() > size.height())
		{
			if(d_display_mode != TAccelerationCalculator::DISPLAY_WIDE)
			{
				d_display_mode = TAccelerationCalculator::DISPLAY_WIDE;
				this->Update_Layout();
			}
		}
		else
		{
			if(d_display_mode != TAccelerationCalculator::DISPLAY_TALL)
			{
				d_display_mode = TAccelerationCalculator::DISPLAY_TALL;
				this->Update_Layout();
			}
		}
		
		this->Update_Graph(size);
	}
}


void TAccelerationCalculator::Update_Layout(void)
{
	d_central_widget_layout->removeItem(d_control_layout);
	d_central_widget_layout->removeWidget(d_graph_label);
	d_central_widget_layout->removeWidget(d_copyright_label);
	
	if(d_display_mode == TAccelerationCalculator::DISPLAY_WIDE)
	{
		d_central_widget_layout->addLayout(d_control_layout,0,0,1,1);
		d_central_widget_layout->addWidget(d_graph_label,0,1,1,1);
		d_central_widget_layout->addWidget(d_copyright_label,1,0,1,2);
	}
	else // if(d_display_mode == TVibrationMeter::DISPLAY_TALL)
	{
		d_central_widget_layout->addLayout(d_control_layout,0,0,1,1);
		d_central_widget_layout->addWidget(d_graph_label,1,0,1,1);
		d_central_widget_layout->addWidget(d_copyright_label,2,0,1,1);
	}

    d_central_widget->updateGeometry();
}

void TAccelerationCalculator::Update_Graph(
	const QSize							&widget_size)
{
	QString								graph_name;
	double								initial_velocity;
	double								final_velocity;
	bool								state;
	static const double					ZERO_EPSILON(0.001);
	
	if(widget_size.width() == 0 ||  widget_size.height() == 0)
	{
		return;
	}
	
	initial_velocity = d_initial_velocity_edit->text().toDouble(&state);
	if(!state) return;
	
	final_velocity = d_final_velocity_edit->text().toDouble(&state);
	if(!state) return;

	if(fabs(initial_velocity - final_velocity) > ZERO_EPSILON)
	{
		if(initial_velocity > final_velocity)
		{
            graph_name = QStringLiteral(":/graphs/graph_down.png");
		}
		else
		{
            graph_name = QStringLiteral(":/graphs/graph_up.png");
		}
	}
	else
	{
        graph_name = QStringLiteral(":/graphs/graph_zero.png");
	}
	
	if(graph_name != d_display_graph_name)
	{
		d_display_graph_name = graph_name;
		d_graph_label->setPixmap(QPixmap(d_display_graph_name));
	}
}



